from typing import Any, Optional

from payton.math.functions import min_max
from payton.math.vector import Vector3D
from payton.scene.geometry.mesh import Mesh
from payton.scene.material import DEFAULT


class Cube(Mesh):
    def __init__(
        self,
        width: float = 1.0,
        depth: float = 1.0,
        height: float = 1.0,
        from_corner: Optional[Vector3D] = None,
        to_corner: Optional[Vector3D] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Cube object

        Instead of using W/D/H arguments, you can define a Cube
        starting from a point A filling up to point B.

        Keyword arguments:
        width -- Width of the cube (default = 1)
        depth -- Depth of the cube (default = 1)
        height -- Height of the cube (default = 1)
        from_corner -- (Optional) Starting point A
        to_corner -- (Optional) Ending point B
        """
        super().__init__(**kwargs)
        width *= 0.5
        depth *= 0.5
        height *= 0.5

        if from_corner is not None and to_corner is not None:
            vmin, vmax = min_max([from_corner, to_corner])

            width = (vmax[0] - vmin[0]) / 2
            depth = (vmax[1] - vmin[1]) / 2
            height = (vmax[2] - vmin[2]) / 2
            self.position = [
                vmin[0] + width,
                vmin[1] + depth,
                vmin[2] + height,
            ]

        _vertices = [
            [width, depth, height],
            [width, depth, -height],
            [width, -depth, height],
            [width, -depth, -height],
            [-width, depth, height],
            [-width, depth, -height],
            [-width, -depth, height],
            [-width, -depth, -height],
        ]

        _normals = [
            [0.0, 0.0, 1.0],
            [0.0, -1.0, 0.0],
            [-1.0, 0.0, 0.0],
            [0.0, 0.0, -1.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ]

        _texcoords = [
            [0.50, 0.333333],
            [0.25, 0.000000],
            [0.25, 0.333333],
            [1.00, 0.333333],
            [0.75, 0.666666],
            [1.00, 0.666666],
            [0.75, 0.333333],
            [0.50, 0.666666],
            [0.25, 0.666666],
            [0.50, 1.000000],
            [0.50, 0.666666],
            [0.25, 0.333333],
            [0.00, 0.666666],
            [0.25, 0.666666],
            [0.50, 0.333333],
            [0.25, 0.666666],
            [0.50, 0.666666],
            [0.50, 0.000000],
            [0.50, 0.333333],
            [0.25, 1.000000],
            [0.00, 0.333333],
            [0.25, 0.333333],
        ]

        _faces = [
            [[4, 0, 0], [2, 1, 0], [0, 2, 0]],
            [[2, 3, 1], [7, 4, 1], [3, 5, 1]],
            [[6, 6, 2], [5, 7, 2], [7, 4, 2]],
            [[1, 8, 3], [7, 9, 3], [5, 10, 3]],
            [[0, 11, 4], [3, 12, 4], [1, 13, 4]],
            [[4, 14, 5], [1, 15, 5], [5, 16, 5]],
            [[4, 0, 0], [6, 17, 0], [2, 1, 0]],
            [[2, 3, 1], [6, 6, 1], [7, 4, 1]],
            [[6, 6, 2], [4, 18, 2], [5, 7, 2]],
            [[1, 8, 3], [3, 19, 3], [7, 9, 3]],
            [[0, 11, 4], [2, 20, 4], [3, 12, 4]],
            [[4, 14, 5], [0, 21, 5], [1, 15, 5]],
        ]

        i = 0
        for index in _faces:
            ind = []
            for f in index:
                l_vertex = _vertices[f[0]]
                l_normal = _normals[f[2]]
                l_tex = _texcoords[f[1]]
                self._vertices.append(l_vertex)
                self._normals.append(l_normal)
                self._texcoords.append(l_tex)
                ind.append(i)
                i += 1
            self.materials[DEFAULT]._indices.append(ind)
        self._indices = self.materials[DEFAULT]._indices

        return None



from typing import List
from payton.math.vector import Vector3D

Matrix = List[Vector3D]

IDENTITY_MATRIX = [
    [1.0, 0.0, 0.0, 0.0],
    [0.0, 1.0, 0.0, 0.0],
    [0.0, 0.0, 1.0, 0.0],
    [0.0, 0.0, 0.0, 1.0],
]



import ctypes
from typing import Any, List, Optional

import numpy as np  # type: ignore
from OpenGL.GL import (
    GL_ARRAY_BUFFER,
    GL_ELEMENT_ARRAY_BUFFER,
    GL_FILL,
    GL_FLOAT,
    GL_FRONT_AND_BACK,
    GL_LINE,
    GL_LINES,
    GL_STATIC_DRAW,
    GL_UNSIGNED_INT,
    glBindBuffer,
    glBindVertexArray,
    glBufferData,
    glDeleteBuffers,
    glDeleteVertexArrays,
    glDrawElements,
    glEnableVertexAttribArray,
    glGenBuffers,
    glGenVertexArrays,
    glIsVertexArray,
    glPolygonMode,
    glVertexAttribPointer,
)

from payton.math.vector import Vector3D
from payton.scene.geometry.base import Line
from payton.scene.material import Material
from payton.scene.shader import Shader


class Grid:
    """Grid is an exceptional object in Payton, which is not extended from
    Object. Instead, it has a simpler structure designed only for the grid
    inside the scene.
    """

    def __init__(
        self,
        xres: int = 20,
        yres: int = 20,
        color: Optional[Vector3D] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Grid

        Keyword arguments:
        xres -- Number of steps in X direction
        yres -- Number of steps in Y direction
        color -- Color of the grid.
        """
        if color is None:
            self._color = [0.4, 0.4, 0.4, 1.0]
        else:
            self._color = color

        self.static: bool = True
        self.matrix: List[float] = [
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
        ]
        self._vertices: List[float] = []
        self._indices: List[int] = []
        self._vertex_count: int = 0
        self._model_matrix: Optional[np.ndarray] = None
        self._material: Material = Material(display=1, lights=False)
        self._material.color = self._color
        self._lines: List[Line] = [
            Line(
                vertices=[
                    [0.0, 0.0, 0.01],
                    [xres / 2.0, 0.0, 0.01],
                ],
                color=[1.0, 0.0, 0.0],
            ),
            Line(
                vertices=[
                    [0.0, 0.0, 0.01],
                    [0.0, yres / 2.0, 0.01],
                ],
                color=[0.0, 1.0, 0.0],
            ),
            Line(
                vertices=[
                    [0.0, 0.0, 0.01],
                    [0.0, 0.0, yres / 2.0],
                ],
                color=[0.0, 0.0, 1.0],
            ),
        ]

        # Vertex Array Object pointer
        self._vao: int = -1
        self.visible: bool = True

        self.resize(xres, yres)

    def destroy(self) -> bool:
        """Destroy the grid object and free the graphics card memory area"""
        if self._vao > -1:
            glDeleteVertexArrays(1, [self._vao])
            self._vao = -1
        for line in self._lines:
            line.destroy()
        return True

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
    ) -> bool:
        """Render the grid"""
        if not self.visible:
            return True

        if self._vao == -1:
            self.build()

        shader.set_matrix4x4_np("model", self._model_matrix)
        self._material.render(False, shader)

        if glIsVertexArray(self._vao):
            glBindVertexArray(self._vao)
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
            glDrawElements(
                GL_LINES,
                self._vertex_count,
                GL_UNSIGNED_INT,
                ctypes.c_void_p(0),
            )
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
            glBindVertexArray(0)

        for line in self._lines:
            line.render(False, shader, None)
        return True

    def resize(self, xres: int, yres: int, spacing: float = 1.0) -> None:
        """Resize the grid

        Keyword arguments:
        xres -- Number of steps in X direction
        yres -- Number of steps in Y direction
        spacing -- Grid spacing
        """
        self._vertices = []
        self._indices = []
        self._vertex_count = 0
        ystart = -(yres * spacing / 2.0)
        xstart = -(xres * spacing / 2.0)
        self._model_matrix = np.asfortranarray(np.array(self.matrix, dtype=np.float32), dtype=np.float32)
        for j in range(0, yres):
            y = ystart + (j * spacing)
            for i in range(0, xres):
                x = xstart + (i * spacing)
                self._vertices += [x, y, 0.0]

        for j in range(0, yres - 1):
            offset = j * xres
            for i in range(0, xres - 1):
                k = offset + i
                self._indices += [
                    k,
                    k + 1,
                    k + 1,
                    k + xres + 1,
                    k + xres + 1,
                    k + xres,
                    k + xres,
                    k,
                ]

        self._vertex_count = len(self._indices)
        if self._vao > -1:
            glDeleteVertexArrays(1, [self._vao])
        self._vao = -1

    @property
    def color(self) -> Vector3D:
        """Return the color of the grid"""
        return self._color

    @color.setter
    def color(self, color: Vector3D) -> None:
        """Set the color of the grid

        Keyword arguments:
        color -- Color of the grid
        """
        self._color = color
        self._material.color = color

    def build(self) -> None:
        """Generate the OpenGL stuff for the grid"""
        self._vao = glGenVertexArrays(1)
        vbos = glGenBuffers(2)
        glBindVertexArray(self._vao)

        vertices = np.array(self._vertices, dtype=np.float32)
        indices = np.array(self._indices, dtype=np.int32)

        glBindBuffer(GL_ARRAY_BUFFER, vbos[0])
        glEnableVertexAttribArray(0)  # shader layout location
        glVertexAttribPointer(0, 3, GL_FLOAT, False, 0, ctypes.c_void_p(0))
        glBufferData(GL_ARRAY_BUFFER, vertices.nbytes, vertices, GL_STATIC_DRAW)

        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vbos[1])
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.nbytes, indices, GL_STATIC_DRAW)
        self._vertex_count = len(indices)

        glBindVertexArray(0)
        # glDisableVertexAttribArray(0)
        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glDeleteBuffers(2, vbos)



from payton.scene.geometry.mesh import Mesh


def export_json(mesh: Mesh, filename: str, **kwargs):
    """Export the given mesh as JSON

    Keyword arguments:
    mesh -- Mesh to export
    filename -- Output filename
    """
    data = mesh.to_json(**kwargs)
    open(filename, "w").write(data)


def import_json(filename: str) -> Mesh:
    """Import the given file as a Mesh

    Keyword arguments:
    filename -- Filename to import"""
    data = open(filename, "r").read()
    return Mesh.from_json(data)



import ctypes
import logging
from typing import Any, Dict, List, Optional

import numpy as np  # type: ignore
from OpenGL.GL import (
    GL_FALSE,
    GL_FRAGMENT_SHADER,
    GL_GEOMETRY_SHADER,
    GL_PROGRAM_POINT_SIZE,
    GL_TRUE,
    GL_VERTEX_SHADER,
    glDisable,
    glEnable,
    glGetUniformLocation,
    glUniform1f,
    glUniform1i,
    glUniform3fv,
    glUniform4fv,
    glUniformMatrix4fv,
    glUseProgram,
    shaders,
)

from payton.math.matrix import Matrix
from payton.math.vector import Vector3D

DEFAULT_SHADER = "default"
PARTICLE_SHADER = "particle"
SHADOW_SHADER = "depth"

depth_fragment_shader = """
#version 330 core
in vec4 FragPos;

uniform float far_plane;
uniform vec3 light_pos[100];

void main()
{
    float lightDistance = length(FragPos.xyz - light_pos[0]);
    lightDistance = lightDistance / far_plane;
    gl_FragDepth = lightDistance;
}
"""

depth_vertex_shader = """
#version 330 core
layout (location = 0) in vec3 aPos;

uniform mat4 model;

void main()
{
    gl_Position = model * vec4(aPos, 1.0);
}
"""

depth_geometry_shader = """
#version 330 core
layout (triangles) in;
layout (triangle_strip, max_vertices=18) out;

uniform mat4 shadowMatrices[6];

out vec4 FragPos; // FragPos from GS (output per emitvertex)

void main()
{
    for(int face = 0; face < 6; ++face)
    {
        gl_Layer = face;
        for(int i = 0; i < 3; ++i) // for each triangle's vertices
        {
            FragPos = gl_in[i].gl_Position;
            gl_Position = shadowMatrices[face] * FragPos;
            EmitVertex();
        }
        EndPrimitive();
    }
}
"""


particle_geometry_shader = """
#version 330 core

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;
uniform int view_mode;
uniform float particle_size;

layout (points) in;
layout (triangle_strip, max_vertices=4) out;

out vec2 varTex;

void main() {
    vec4 center = view * model * gl_in[0].gl_Position;
    varTex = vec2( 0, 0 );
    gl_Position = projection * (center + vec4(particle_size, particle_size, 0, 0));
    EmitVertex();

    varTex = vec2( 1, 0  );
    gl_Position = projection * (center + vec4(particle_size, -particle_size, 0, 0));
    EmitVertex();

    varTex = vec2( 0, 1 );
    gl_Position = projection * (center + vec4(-particle_size, particle_size, 0, 0));
    EmitVertex();

    varTex = vec2( 1, 1 );
    gl_Position = projection * (center + vec4(-particle_size, -particle_size, 0, 0));
    EmitVertex();
}

"""


particle_fragment_shader = """
#version 330

uniform sampler2D tex_unit;
uniform vec3 object_color;
uniform float opacity;

in vec2 varTex;

layout(location = 0) out vec4 outFragColor;

void main()
{
    vec4 tex_color;
    tex_color = texture( tex_unit, varTex );

    outFragColor = tex_color * vec4(object_color, opacity);
}
"""


particle_vertex_shader = """
#version 330 core
layout ( location = 0 ) in vec3 position;
layout ( location = 1 ) in vec3 normal;
layout ( location = 2 ) in vec2 texCoords;
layout ( location = 3 ) in vec3 colors; // optional

out vec2 tex_coords;
out vec3 frag_color;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;
uniform int view_mode;

out vec3 l_fragpos;
out vec3 l_normal;

void main()
{
    gl_Position = vec4(position, 1.0f);
}
"""  # type: str


default_fragment_shader = """
#version 330 core
out vec4 FragColor;

in vec2 tex_coords;
in vec3 frag_color;

in vec3 l_fragpos;
in vec3 l_normal;

uniform vec3 light_pos[100]; // assume 100 lights max.
uniform vec3 light_color[100];
uniform vec3 camera_pos;
uniform bool shadow_enabled;
uniform int samples;

uniform int LIGHT_COUNT;

uniform vec3 object_color;
uniform int material_mode;
uniform float opacity;
uniform float far_plane;
uniform int lit;

uniform sampler2D tex_unit;
uniform samplerCube depthMap;

// array of offset direction for sampling
vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1, 1,  1),
   vec3(1, 1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1, 1,  0),
   vec3(1, 0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1, 0, -1),
   vec3(0, 1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0, 1, -1)
);

float ShadowCalculation(vec3 fragPos)
{
    vec3 fragToLight = fragPos - light_pos[0];
    float currentDepth = length(fragToLight);
    float shadow = 0.0;
    float bias = 0.15;

    float viewDistance = length(camera_pos - fragPos);
    float diskRadius = (1.0 + (viewDistance / far_plane)) / 25.0;

    for(int i = 0; i < samples; ++i)
    {
        float closestDepth = texture(depthMap, fragToLight +
                                     gridSamplingDisk[i] * diskRadius).r;
        closestDepth *= far_plane;
        if(currentDepth - bias > closestDepth)
             shadow += 1.0;
    }

    return shadow / float(samples);
}

void main()
{
    vec3 color;
    float alpha;
    if (material_mode == 0 || material_mode == 2) {
        color = object_color;
        alpha = opacity;
    }
    if (material_mode == 1 || material_mode == 3) {
        color = texture(tex_unit, tex_coords).rgb;
        alpha = texture(tex_unit, tex_coords).a;
    }
    if (material_mode == 4) {
        color = frag_color;
        alpha = opacity;
    }

    if (lit == 0) {
        FragColor = vec4(color, alpha);
    } else {
        vec3 norm = normalize(l_normal);
        for (int i = 0; i < LIGHT_COUNT; i++) {
            vec3 ambient = 0.3 * light_color[i];

            // diffuse
            vec3 norm = normalize(l_normal);
            vec3 lightDir = normalize(light_pos[i] - l_fragpos);
            float diff = max(dot(norm, lightDir), 0.0);
            vec3 diffuse = diff * light_color[i];
            float spec = 0.0;
            vec3 halfwayDir = normalize(lightDir + camera_pos);
            spec = pow(max(dot(norm, halfwayDir), 0.0), 64.0);

            vec3 specular = spec * light_color[i];
            float shadow = shadow_enabled ? ShadowCalculation(l_fragpos) : 0.0;
            vec3 lighting = (ambient + (1.0 - shadow) *
                             (diffuse + specular)) * color;

            FragColor = vec4(lighting, alpha);
        }
    }
}"""  # type: str


default_vertex_shader = """
#version 330 core
layout ( location = 0 ) in vec3 position;
layout ( location = 1 ) in vec3 normal;
layout ( location = 2 ) in vec2 texCoords;
layout ( location = 3 ) in vec3 colors; // optional

out vec2 tex_coords;
out vec3 frag_color;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;
uniform int view_mode;

out vec3 l_fragpos;
out vec3 l_normal;

void main()
{
    l_fragpos = vec3(model * vec4(position, 1.0));
    l_normal = mat3(transpose(inverse(model))) * normal;

    if (view_mode == 0) {
        gl_Position = projection * view * model * vec4(position, 1.0f);
    }
    if (view_mode == 1) {
        gl_Position = (projection * (vec4(position, 1.0f)
                        + vec4(model[3].xyz, 0.0f)));
    }

    gl_PointSize = 5.0;

    tex_coords = texCoords;
    frag_color = colors;
}
"""  # type: str


background_vertex_shader = """
#version 330 core
out vec2 v_uv;

void main()
{
  // uint idx = gl_VertexID;
  gl_Position = (vec4( gl_VertexID & 1, gl_VertexID >> 1, 0.0, 0.5 )
                 * 4.0 - 1.0);
  v_uv = vec2( gl_Position.xy * 0.5 + 0.5 );
}
"""  # type: str

background_fragment_shader = """
#version 330 core
uniform vec4 top_color;
uniform vec4 bot_color;
in vec2 v_uv;

out vec4 frag_color;
void main()
{
  frag_color = top_color * (1 - v_uv.y) + bot_color * v_uv.y;
}
"""  # type: str


class Shader:
    NO_LIGHT_COLOR = 0  # type: int
    NO_LIGHT_TEXTURE = 1  # type: int
    LIGHT_COLOR = 2  # type: int
    LIGHT_TEXTURE = 3  # type: int
    PER_VERTEX_COLOR = 4  # type:int
    HUD = 5  # type: int
    DEPTH_CALC = 6  # type: int

    def __init__(
        self,
        fragment: str = default_fragment_shader,
        vertex: str = default_vertex_shader,
        geometry: str = "",
        variables: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        """Initialize the shader

        Keyword arguments:
        fragment -- Fragment shader code
        vertex -- Vertex shader code
        geometry -- Geometry shader code
        variables -- Shader variables (Optional but defining them beforehand can increase the warm-up performance)
        """
        self.fragment_shader_source = fragment
        self.vertex_shader_source = vertex
        self.geometry_shader_source = geometry

        self.variables: List[str] = [] if variables is None else variables
        self._stack: Dict[str, int] = {}  # Variable stack.
        self._mode: int = self.NO_LIGHT_COLOR  # Lightless color material
        self._depth_shader = False

        self.program: int = -1

    def build(self) -> int:
        """Build the shader

        Compile the shader codes and compile the program
        """
        vertex_shader = shaders.compileShader(self.vertex_shader_source, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.fragment_shader_source, GL_FRAGMENT_SHADER)
        geometry_shader = None
        if self.geometry_shader_source != "":
            geometry_shader = shaders.compileShader(self.geometry_shader_source, GL_GEOMETRY_SHADER)

        if geometry_shader is not None:
            self.program = shaders.compileProgram(vertex_shader, fragment_shader, geometry_shader, validate=False)
        else:
            self.program = shaders.compileProgram(vertex_shader, fragment_shader, validate=False)

        return self.program

    def use(self) -> bool:
        """Use the shader (activate it in the pipeline)"""
        if self.program == -1:
            logging.error("Shader not compiled")
            return False
        glUseProgram(self.program)
        glEnable(GL_PROGRAM_POINT_SIZE)
        return True

    def end(self) -> None:
        """End using the shader in the pipeline"""
        glUseProgram(0)
        glDisable(GL_PROGRAM_POINT_SIZE)

    def set_matrix4x4_np(self, variable: str, value: np.ndarray, transpose: bool = False) -> bool:
        """Set 4x4 Matrix data in the shader as Numpy Array

        Keyword arguments:
        variable -- Variable name
        value -- Numpy array as the value
        transpose -- GL_TRANSPOSE variable if needed (glUniformMatrix4fv)
        """
        g_transpose = GL_TRUE if transpose else GL_FALSE
        location = self.get_location(variable)
        if location == -1:
            return False

        glUniformMatrix4fv(
            location,
            1,
            g_transpose,
            value,
        )
        return True

    def get_location(self, variable: str) -> int:
        """Get the memory location of the given variable in the shader

        Keyword arguments:
        variable -- Variable name
        """
        if variable in self._stack:
            return self._stack[variable]
        location = glGetUniformLocation(self.program, variable)

        self._stack[variable] = location
        return location

    def set_matrix4x4(self, variable: str, value: Matrix, transpose: bool = False) -> None:
        """Set 4x4 Matrix data in the shader

        Keyword arguments:
        variable -- Variable name
        value -- Matrix value
        transpose -- GL_TRANSPOSE variable if needed (glUniformMatrix4fv)
        """
        self.set_matrix4x4_np(variable, np.array(value, np.float32), transpose)

    def set_vector3_array_np(self, variable: str, value: np.ndarray, count: int) -> bool:
        """Set the given vectors as value in the shader

        Keyword arguments:
        variable -- Variable name
        value -- Value as numpy array
        count -- Number of elements
        """
        value = value.flatten()
        location = self.get_location(variable)
        if location == -1:
            return False
        glUniform3fv(location, count, value[:3])
        return True

    def set_vector3_np(self, variable: str, value: np.ndarray) -> bool:
        """Set the given vector value in the shader

        Keyword arguments:
        variable -- Variable name
        value -- Numpy array value
        """
        location = self.get_location(variable)
        if location == -1:
            return False
        glUniform3fv(location, 1, value[:3])
        return True

    def set_vector4_np(self, variable: str, value: np.ndarray) -> bool:
        """Set the given vector as 4 elements in the shader

        Keyword arguments:
        variable -- Variable name
        value -- Numpy array value
        """
        location = self.get_location(variable)
        if location == -1:
            return False
        glUniform4fv(location, 1, value)
        return True

    def set_vector3(self, variable: str, value: Vector3D) -> None:
        """Set the given vector in the shader

        Keyword arguments:
        variable -- Variable name
        value -- Value as Vector3D
        """
        self.set_vector3_np(variable, np.array(list(value), dtype=np.float32))

    def set_int(self, variable: str, value: int) -> bool:
        """Set integer in the shader

        Keyword arguments:
        variable -- Variable name
        value -- Value as integer
        """
        location = self.get_location(variable)
        if location == -1:
            return False

        glUniform1i(location, ctypes.c_int(value))
        return True

    def set_float(self, variable: str, value: float) -> bool:
        """Set float in the shader

        Keyword arguments:
        variable -- Variable name
        value -- Value as float
        """
        location = self.get_location(variable)
        if location == -1:
            return False

        glUniform1f(location, ctypes.c_float(value))
        return True



import math
import os
from payton.math.vector import Vector3D
from textwrap import wrap
from typing import Any, Callable, Dict, Optional, List, Tuple, TypeVar, cast

import numpy as np  # type: ignore
from OpenGL.GL import GL_DEPTH_TEST, glDisable, glEnable
from PIL import Image, ImageDraw, ImageFont  # type: ignore

from payton.math.functions import ortho
from payton.scene.geometry.base import Object
from payton.scene.geometry.mesh import Mesh
from payton.scene.shader import Shader

S2 = TypeVar("S2", bound="Shape2D")


class Shape2D(Mesh):
    def __init__(
        self,
        position: List[int],
        size: List[int],
        on_click: Optional[Callable] = None,
        opacity: float = 0.1,
        **kwargs: Any,
    ):
        """Initialize 2D Shape

        Keyword arguments:
        position -- Position of the shape on the screen (x, y, and optionally Z as integer)
        size -- Size of the shape (w, h as integer)
        on_click -- On Click event callback (takes no arguments)
        opacity -- Opacity of the shape
        """
        super().__init__(**kwargs)
        self.material.opacity = opacity
        position = list(position)
        if len(position) == 2:
            position.append(0)
        position[2] = -1
        self.__position = position
        self.position = list([float(x) for x in self.__position])
        self.size = size
        self.on_click: Optional[Callable] = on_click
        self._font: ImageFont = None
        self.parent: Any = None
        self._scene_width: int = 0
        self._scene_height: int = 0

    def add_child(self, name: str, obj: "Shape2D") -> bool:  # type: ignore
        """Add a child shape to this shape object.

        This method overrides the parent by arguments. Child object's position
        follows the parent.

        Keyword arguments:
        name -- Name of the object to be added
        obj -- Shape2D Object to add.
        """
        res = super().add_child(name, obj)  # type: ignore
        if not res:
            return res
        obj.parent = self
        return res

    def draw(self):
        """Placeholder for draw function"""
        pass

    def draw_text(self):
        """Placeholder for draw text function"""
        pass

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: int = None,
    ):
        """Render cycle for the Shape 2D

        We override the `lit` parameter for the super.
        This is not actually intended to be used publicly except if you want to
        setup your own render cycle.
        """
        super().render(False, shader, parent_matrix)

    @property
    def font(self) -> None:
        """Return the fond of the shape"""
        if self._font is not None:
            return self._font
        if self.parent is not None:
            return self.parent.font
        return None

    @font.setter
    def font(self, font: ImageFont) -> None:
        """Set the font of the Shape

        Keyword arguments:
        font -- ImageFont object
        """
        self._font = font
        self.draw_text()

    def _set_parent_size(self, w: int, h: int):
        self._parent_width = w
        self._parent_height = h
        self._init = False
        self.draw()
        for child in self.children.values():
            cast("Shape2D", child)._set_parent_size(self.size[0], self.size[1])

    def click(self, x: int, y: int) -> Optional[Mesh]:
        """Check for click event

        Keyword arguments:
        x -- Cursor X
        y -- Cursor Y
        """
        if not callable(self.on_click):
            for child in self.children:
                c = cast(Mesh, self.children[child]).click(x, y)
                if c:
                    return cast(Mesh, self.children[child])
            return None

        if self._model_matrix is None:
            return None
        mm = self._model_matrix[3]
        if x > mm[0] and x < mm[0] + self.size[0] and y > mm[1] and y < mm[1] + self.size[1]:
            self.on_click()
            return self
        return None


class Rectangle(Shape2D):
    def __init__(
        self,
        position: List[int],
        size: List[int],
        **kwargs: Any,
    ):
        """Initialize Rectangle

        Keyword arguments:
        position -- Tuple as [X, Y, Z] positions
        size -- Size of the rectangle
        """
        super().__init__(position=position, size=size, **kwargs)
        self._init: bool = False
        self.draw()

    def draw(self) -> None:
        """Create the rectangle polygons"""
        w, h = self.size
        self.clear_triangles()
        if not self._init:
            self.add_triangle(
                [[0, 0, 1], [w, h, 1], [w, 0, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
            )
            self.add_triangle(
                [[0, 0, 1], [0, h, 1], [w, h, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
            )
            self._init = True


class Text(Rectangle):
    def __init__(
        self,
        position: List[int],
        size: List[int],
        label: str = "lorem",
        bgcolor: Optional[Vector3D] = None,
        color: Optional[Vector3D] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize Text object (label)

        Keyword arguments:
        position -- Tuple as [X, Y, Z] positions on the screen
        size -- Size of the text area
        label -- Text / Label to be printed
        bgcolor -- Background color of the text
        color -- Color of the text"""
        super().__init__(position=position, size=size, **kwargs)
        self.__label: str = label
        self.bgcolor: Vector3D = [0, 0, 0, 0]
        if bgcolor is not None:
            self.bgcolor = bgcolor

        self.color: Vector3D = [0, 0, 0]
        if color is not None:
            self.color = color

        self.crop = [0, 0, 0, 0]
        self._init_text: bool = False
        self._max_char_width = 0.0

    @property
    def label(self) -> str:
        """Get the label of the text"""
        return self.__label

    @label.setter
    def label(self, label: str) -> None:
        """Set the label of the text

        Keyword arguments:
        label -- Label string
        """
        self.__label = label
        self._init_text = False

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: int = None,
    ) -> None:
        """Render the text object by initializing the text material if not initialized yet

        Keyword arguments:
        lit -- Is the object illuminated?
        shader -- Shader to use while rendering the object
        parent_matrix -- Parent object's matrix if this is a child object
        _primitive -- override the scene wide primitive (rendering) mode. (Point, Wire, Solid)
        """
        if not self._init_text:
            self.draw_text()
        super().render(lit, shader, parent_matrix)

    @property
    def text_size(self) -> Tuple[int, int]:
        """Return the text size in pixels"""
        res = (0, 0)
        timg = Image.new("RGBA", (1, 1))
        d = ImageDraw.Draw(timg)

        res = d.textsize(self.label, font=self.font)
        lres = list(res)
        if lres[0] < self.size[0]:
            lres[0] = self.size[0]
        if lres[1] < self.size[1]:
            lres[1] = self.size[1]
        return lres[0], lres[1]

    def wrap(self, width_in_pixels: int):
        """Word-wrap the text to fit into the given pixel size

        Keyword arguments:
        width_in_pixels -- Width size in pixels to fit the text into
        """
        original = self.label
        self.label = ""
        if self.font is None:
            return
        if self._max_char_width == 0.0:
            self._max_char_width = self.font.getsize('_')[0]
        split_length = int(math.floor(width_in_pixels / self._max_char_width)) - 1
        parts = wrap(original, split_length)
        self.__label = "\n".join(parts)

        self._init_text = False

    def draw_text(self) -> None:
        """Create the text material and initialize the font if needed"""
        if self._init_text:
            return
        bgcolor = (
            int(self.bgcolor[0] * 255),
            int(self.bgcolor[1] * 255),
            int(self.bgcolor[2] * 255),
            int(self.bgcolor[3] * 255),
        )
        img = Image.new("RGBA", self.text_size, color=bgcolor)
        d = ImageDraw.Draw(img, "RGBA")
        color = (int(self.color[0] * 255), int(self.color[1] * 255), int(self.color[2] * 255), 255)
        if self.font is not None:
            d.text((1, 1), self.label, fill=color, font=self.font)
        else:
            d.text((1, 1), self.label, fill=color)
        if any(self.crop):
            img = img.crop(self.crop)
        del d

        self.material._image = img
        self.material.opacity = 1.0
        self.material.refresh()
        self._init_text = True


class Hud(Object):
    def __init__(
        self,
        width: int = 800,
        height: int = 600,
        font: str = "",
        font_size: int = 15,
        **kwargs: Any,
    ) -> None:
        """Initialize HUD Object to render HUD elements.

        You need a HUD in a scene to render 2D shapes on top of your scene

        Keyword arguments:
        width -- Width of the HUD area
        height -- Height of the HUD area
        font -- Font name to use (default: monofonto provided by the package)
        font_size -- Font size in pixels
        """
        super().__init__(**kwargs)
        self.width: int = width
        self.height: int = height
        self._fontname: str = font
        self.children: Dict[str, Shape2D] = {}  # type: ignore
        self._font_size: int = font_size
        if self._fontname != "":
            self.set_font(self._fontname, self._font_size)
        self._font: ImageFont = None
        self._projection_matrix: Optional[np.ndarray] = None
        try:
            self.set_font(os.path.join(os.path.dirname(os.path.abspath(__file__)), "monofonto.ttf"))
        except OSError:
            # font not found but pillow has better than nothinf font
            pass

    @property
    def font(self) -> ImageFont:
        """Return the HUD Image Font"""
        return self._font

    def add_child(self, name: str, obj: Shape2D) -> bool:  # type: ignore
        """Add 2D Shape into Hud.

        Note: this is a type ignore due to it's mismatch with Object add_child method

        Keyword arguments:
        name -- Name of the object
        obj -- Shape2D object to add"""
        res = super().add_child(name, obj)
        obj._scene_height = self.height
        obj._scene_width = self.width

        if not res:
            return res
        obj.parent = self
        return res

    def render(self, lit: bool, shader: Shader, *_args: Any) -> None:
        """Main render cycle for HUD

        Disables the DEPTH test and draws the objects from parent to child
        """
        if not self.visible:
            return
        lit = False
        glDisable(GL_DEPTH_TEST)
        if self._projection_matrix is None:
            self._projection_matrix = ortho(0, self.width, self.height, 0)

        shader.set_int("view_mode", 1)
        shader.set_matrix4x4_np("projection", self._projection_matrix)
        for child in self.children:
            self.children[child].render(lit, shader, None)
        glEnable(GL_DEPTH_TEST)

    def set_size(self, w: int, h: int):
        """Set HUD size

        Keyword arguments:
        w -- Width of the HUD
        h -- Height of the HUD
        """
        self.width = w
        self.height = h
        for child in self.children.values():
            cast(Shape2D, child)._set_parent_size(w, h)
            child._init = False
            child.draw()

        self._projection_matrix = None

    def set_font(self, font_name: str, font_size: int = 16) -> None:
        """Set True Type font for the Hud

        Keyword arguments:
        font_name -- Name of the True Type font installed in the system
        font_size -- Size of the font in pixels
        """
        self._font = ImageFont.truetype(font_name, font_size)



import os
from typing import Any, Dict, List, Optional

import numpy as np  # type: ignore
from OpenGL.GL import (
    GL_BLEND,
    GL_CULL_FACE,
    GL_LINEAR,
    GL_LINEAR_MIPMAP_LINEAR,
    GL_ONE,
    GL_ONE_MINUS_SRC_ALPHA,
    GL_REPEAT,
    GL_RGB,
    GL_RGBA,
    GL_SRC_ALPHA,
    GL_TEXTURE0,
    GL_TEXTURE_2D,
    GL_TEXTURE_MAG_FILTER,
    GL_TEXTURE_MIN_FILTER,
    GL_TEXTURE_WRAP_S,
    GL_TEXTURE_WRAP_T,
    GL_UNPACK_ALIGNMENT,
    GL_UNSIGNED_BYTE,
    glActiveTexture,
    glBindTexture,
    glBlendFunc,
    glDisable,
    glEnable,
    glGenerateMipmap,
    glGenTextures,
    glPixelStorei,
    glTexImage2D,
    glTexParameterf,
)
from PIL import Image  # type: ignore

from payton.math.vector import Vector3D
from payton.scene.shader import Shader
from payton.scene.types import IList

SOLID = 0  # type: int
WIREFRAME = 1  # type: int
POINTS = 2  # type: int

RED = [1.0, 0.0, 0.0]  # type: Vector3D
GREEN = [0.0, 1.0, 0.0]  # type: Vector3D
BLUE = [0.0, 0.0, 1.0]  # type: Vector3D
CRIMSON = [220 / 255.0, 20 / 255.0, 60 / 255.0]  # type: Vector3D
PINK = [1.0, 192 / 255.0, 203 / 255.0]  # type: Vector3D
VIOLET_RED = [1.0, 62 / 255.0, 150 / 255.0]  # type: Vector3D
DEEP_PINK = [1.0, 20 / 255.0, 147 / 255.0]  # type: Vector3D
ORCHID = [218 / 255.0, 112 / 255.0, 214 / 255.0]  # type: Vector3D
PURPLE = [128 / 255.0, 0.0, 128 / 255.0]  # type: Vector3D
NAVY = [0.0, 0.0, 0.5]  # type: Vector3D
ROYAL_BLUE = [65 / 255.0, 105 / 255.0, 225 / 255.0]  # type: Vector3D
LIGHT_STEEL_BLUE = [176 / 255.0, 196 / 255.0, 222 / 255.0]  # type: Vector3D
STEEL_BLUE = [70 / 255.0, 130 / 255.0, 180 / 255.0]  # type: Vector3D
TURQUOISE = [0.0, 245 / 255.0, 1.0]  # type: Vector3D
YELLOW = [1.0, 1.0, 0.0]  # type: Vector3D
GOLD = [1.0, 225 / 255.0, 0.0]  # type: Vector3D
ORANGE = [1.0, 165 / 255.0, 0.0]  # type: Vector3D
WHITE = [1.0, 1.0, 1.0]  # type: Vector3D
BLACK = [0.0, 0.0, 0.0]  # type: Vector3D
DARK_GRAY = [0.2, 0.2, 0.2]  # type: Vector3D
LIGHT_GRAY = [0.8, 0.8, 0.8]  # type: Vector3D

DEFAULT = "default"
NO_VERTEX_ARRAY = -1
NO_INDICE = -2
EMPTY_VERTEX_ARRAY = -3

BASE_PARTICLE = os.path.join(os.path.dirname(os.path.abspath(__file__)), "particle.png")


class Material:
    def __init__(
        self,
        color: Optional[Vector3D] = None,
        display: int = SOLID,
        lights: bool = True,
        texture: str = "",
        opacity: float = 1.0,
        **kwargs: Any,
    ):
        """Payton materials are quite simple and does not support some
        functionalities like Game Engines or design softwares.

        Keyword arguments:
        color -- Material color
        display -- Display type of the material (SOLID, WIREFRAME, POINTS)
        lights -- Is this material effected by the light shading?
        texture -- Texture filename
        opacity -- Opacity of the material
        """
        self._color = [1.0, 1.0, 1.0] if color is None else color
        self._color_np: np.ndarray = np.array(list(self._color), dtype=np.float32)
        self.display: int = display
        self.lights: bool = lights
        self.texture: str = texture
        self.particle_texture: str = BASE_PARTICLE
        self.opacity: float = opacity
        self.particle_size: float = 0.16
        self._image: Optional[Image.Image] = None
        self._indices: IList = []
        self._vao: int = NO_VERTEX_ARRAY
        self._vbos: List[int] = []

        self._vertex_count: int = 0
        self._index_count: int = 0

        self._initialized: bool = False
        self._texture: Optional[int] = None
        self._particle_texture: Optional[int] = None

    def to_dict(self) -> Dict[str, Any]:
        """Convert the material into dictionary"""
        return {
            "color": self.color,
            "display": self.display,
            "texture": self.texture,
            "opacity": self.opacity,
            "indices": self._indices,
        }

    @property
    def index_count(self) -> int:
        """Return the number of indexes for OpenGL"""
        if self._index_count > 0:
            return self._index_count
        self._index_count = len(self._indices)
        return self._index_count

    @property
    def color(self) -> Vector3D:
        """Return the material color"""
        return self._color

    @color.setter
    def color(self, color: Vector3D) -> None:
        """Set the material color

        Keyword arguments:
        color -- Color to set
        """
        self._color = color
        self._color_np = np.array(list(self._color), dtype=np.float32)

    @classmethod
    def from_dict(cls, material_dictionary: Dict[str, Any]) -> "Material":
        """Import material from dictionary

        material_dictionary -- Dictionary to import"""
        res = cls()
        res.color = material_dictionary["color"]
        res.display = material_dictionary["display"]
        res.texture = material_dictionary["texture"]
        res.opacity = material_dictionary["opacity"]
        res._indices = material_dictionary["indices"]
        return res

    def build(self) -> bool:
        """Build the material"""
        self._initialized = True
        if os.path.isfile(self.texture):
            img = Image.open(self.texture)
            self.load_texture(img)
        if self._image is not None:
            self.load_texture(self._image)
        if os.path.isfile(self.particle_texture):
            img = Image.open(self.particle_texture)
            self.load_texture(img, particle=True)
        return True

    def load_texture(self, img: Image.Image, particle: bool = False) -> None:
        """Load texture directly from PIL Image object

        Keyword arguments:
        img -- Image to load
        particle -- Is this a particle material?
        """
        img_data = np.fromstring(img.tobytes(), np.uint8)
        width, height = img.size
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
        if particle:
            self._particle_texture = glGenTextures(1)
            glBindTexture(GL_TEXTURE_2D, self._particle_texture)
        else:
            self._texture = glGenTextures(1)
            glBindTexture(GL_TEXTURE_2D, self._texture)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)

        mode = GL_RGBA
        if img.mode == "RGB":
            mode = GL_RGB
        if img.mode == "P":
            img = img.convert("RGB")
            img_data = np.fromstring(img.tobytes(), np.uint8)
            mode = GL_RGB
        glTexImage2D(
            GL_TEXTURE_2D,
            0,
            mode,
            width,
            height,
            0,
            mode,
            GL_UNSIGNED_BYTE,
            img_data,
        )
        glGenerateMipmap(GL_TEXTURE_2D)
        glBindTexture(GL_TEXTURE_2D, 0)

    def refresh(self) -> None:
        """Refresh / apply the material changes into OpenGL context"""
        self._initialized = False

    def material_mode(self, lit: bool) -> int:
        """Return the material mode

        Keyword argument:
        lit -- Is this a lit material?
        """
        mode = Shader.LIGHT_COLOR

        if self.display == SOLID:
            if lit and self.lights:
                if self._texture is not None:
                    mode = Shader.LIGHT_TEXTURE
                else:
                    mode = Shader.LIGHT_COLOR
            else:
                if self._texture is not None:
                    mode = Shader.NO_LIGHT_TEXTURE
                else:
                    mode = Shader.NO_LIGHT_COLOR
        else:
            mode = Shader.NO_LIGHT_COLOR
        return mode

    def render(
        self,
        lit: bool,
        shader: Shader,
        mode: Optional[int] = None,
    ) -> None:
        """Render the material

        Keyword arguments:
        lit -- Is this a lit material?
        shader -- Shader to use for rendering the material
        mode -- Material mode
        """
        if not self._initialized:
            self.build()

        _mode = mode if mode else self.material_mode(lit)

        glEnable(GL_BLEND)
        glDisable(GL_CULL_FACE)
        blend = GL_ONE_MINUS_SRC_ALPHA
        if self.display == POINTS:
            blend = GL_ONE
        glBlendFunc(GL_SRC_ALPHA, blend)

        if self._texture is not None and self.display != POINTS:
            check = shader.get_location("tex_unit")
            if check > -1:
                glActiveTexture(GL_TEXTURE0)
                glBindTexture(GL_TEXTURE_2D, self._texture)
                shader.set_int("tex_unit", 0)

        if self._particle_texture is not None and self.display == POINTS:
            check = shader.get_location("tex_unit")
            shader.set_float("particle_size", self.particle_size)
            if check > -1:
                glActiveTexture(GL_TEXTURE0)
                glBindTexture(GL_TEXTURE_2D, self._particle_texture)
                shader.set_int("tex_unit", 0)

        if not shader._depth_shader:
            shader.set_vector3_np("object_color", self._color_np)
            shader.set_float("opacity", self.opacity)
            shader.set_int("material_mode", _mode)
            shader.set_int("lit", 1 if lit else 0)
            if not self.lights:
                shader.set_int("lit", 0)



"""Payton Line to Mesh module"""
from typing import List

from payton.math.functions import add_vectors, create_rotation_matrix_raw, scale_vector, vector_transform
from payton.math.vector import Vector3D
from payton.scene.geometry import Line, Mesh


def extrude_line(line: Line, direction: Vector3D, distance: float) -> Mesh:
    """Extrude the given line in given direction to form a mesh.

    As a simple example, you can define the cross-section of a wall
    as a Line and extrude it in [0, 0, 1] direction to turn it into a 3D wall.

    Keyword arguments:
    line -- Cross-Section Line object
    direction -- Direction to extrude the given line
    distance -- Distance/Height to extrude
    """
    vertices = line._vertices
    diff_vector = scale_vector(direction, distance)
    mirror_vertices = [add_vectors(v, diff_vector) for v in vertices]

    mesh = Mesh()
    for i in range(len(vertices) - 1):
        mesh.add_triangle(
            vertices=[vertices[i], vertices[i + 1], mirror_vertices[i]],
            texcoords=[[0.0, 0.0], [1.0, 0.0], [0.0, 1.0]],
        )
        mesh.add_triangle(
            vertices=[vertices[i + 1], mirror_vertices[i + 1], mirror_vertices[i]],
            texcoords=[[1.0, 0.0], [1.0, 1.0], [0.0, 1.0]],
        )
    return mesh


def rotate_line(line: Line, axis: Vector3D, angle: float, steps: int = 10) -> Mesh:
    """Rotate the given line around the given axis by the given angle and steps to
    form a mesh.

    As an example, you can draw the cross section of a vase and rotate it around [0, 0, 1]
    by 2PI radians to form a vase.

    Keyword arguments:
    line -- Line object to rotate
    axis -- Rotation axis
    angle -- Angle of rotation in radians
    steps -- Steps to rotate
    """
    step_angle = angle / steps
    vertices = line._vertices
    step_u = 1.0 / steps
    step_v = 1.0 / len(vertices)
    mesh = Mesh()

    for i in range(steps):
        matrix = create_rotation_matrix_raw(axis, step_angle)
        mirror_vertices = [vector_transform(v, matrix) for v in vertices]
        for j in range(len(vertices) - 1):
            mesh.add_triangle(
                vertices=[vertices[j], vertices[j + 1], mirror_vertices[j]],
                texcoords=[[step_u * i, step_v * j], [step_u * (i + 1), step_v * j], [step_u * i, step_v * (j + 1)]],
            )
            mesh.add_triangle(
                vertices=[vertices[j + 1], mirror_vertices[j + 1], mirror_vertices[j]],
                texcoords=[
                    [step_u * (i + 1), step_v * j],
                    [step_u * (i + 1), step_v * (j + 1)],
                    [step_u * i, step_v * (j + 1)],
                ],
            )
        vertices = mirror_vertices.copy()
    mesh.fix_normals()
    return mesh


def lines_to_mesh(lines: List[Line]) -> Mesh:
    """Fill the gaps between several lines with the same number of vertices to form a mesh

    Assume that you have two lines and you want to form a wall between two of them

    Keyword arguments:
    lines -- Line objects to fill between
    """
    lens = [len(line._vertices) for line in lines]
    lmin = min(lens)
    lmax = max(lens)
    mesh = Mesh()
    if lmin != lmax:
        raise BaseException("Number of vertices for each line must be equal")
    if len(lines) == 0:
        raise BaseException("You must specify more than one line object")
    for i in range(len(lines) - 1):
        fvlist = lines[i]._vertices
        tvlist = lines[i + 1]._vertices
        for j in range(lmin - 1):
            mesh.add_triangle(vertices=[fvlist[j], fvlist[j + 1], tvlist[j]])
            mesh.add_triangle(vertices=[fvlist[j + 1], tvlist[j + 1], tvlist[j]])
    mesh.fix_normals()
    mesh.fix_texcoords()
    return mesh



import logging
from itertools import combinations
from typing import Any, Callable, List, Optional, Set

from payton.math.geometry import distance
from payton.scene.geometry.base import Object
from payton.scene.geometry.mesh import Mesh
from payton.scene.geometry.sphere import Sphere


class Collision:
    """Collision pair"""

    def __init__(self, obj1: Object, obj2: Object) -> None:
        """Collision pair

        obj1 -- Object 1
        obj2 -- Object 2
        """
        self.object_1 = obj1
        self.object_2 = obj2
        self.segment = 1


class CollisionTest:
    """
    Main collision test

    Currently, there are two types of collision test algorithms

    AABB -- Axis Aligned Bounding Box intersection test
    SPHERICAL -- Bounding Sphere intersection test

    Both of the algorithms are just simple assumptions and not
    accurate. But any better algorithms are currently hard to
    implement without a big performance impact. At least, I tried.
    Any contributions are welcome.

    TODO: Improve collision detection algorithms.
    """

    SPHERICAL = 0
    AABB = 1

    def __init__(
        self,
        callback: Callable,
        level: int = AABB,
        objects: Optional[List[Mesh]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the test case.

        Keyword arguments:
        callback -- Callback method to call in case of an intersection
                    (Eg: def on_collide(test_case: CollisionTest, collision_pairs: List[Collision] = []))
        level -- Level of the test (AABB, SPHERICAL, default AABB)
        objects -- List of objects inside the test case
        """
        self.objects: List[Mesh] = [] if objects is None else objects
        self.callback: Callable = callback
        self.level: int = level
        self._pairs: List[Set[Mesh]] = []

    def add_object(self, obj: Mesh) -> None:
        """Add a mesh object to the test case

        Keyword arguments:
        obj -- Object to add"""
        if not isinstance(obj, Mesh):
            logging.error("object must be an instance of Mesh")
            return
        self.objects.append(obj)

    def _dist(self, obj1: Object, obj2: Object) -> float:
        p1 = obj1._model_matrix[3]
        p2 = obj2._model_matrix[3]
        return distance(p1, p2)

    def _bounding_sphere_collision(self, obj1: Object, obj2: Object) -> bool:
        dist = self._dist(obj1, obj2)
        total_radius = obj1.bounding_radius + obj2.bounding_radius
        return dist <= total_radius

    def _sphere_in_sphere_collision(self, obj1: Mesh, obj2: Mesh) -> bool:
        dist = self._dist(obj1, obj2)
        if obj1.bounding_radius > dist + obj2.bounding_radius:
            return True
        if obj2.bounding_radius > dist + obj1.bounding_radius:
            return True
        return False

    def _aabb_collision_test(self, obj1: Mesh, obj2: Mesh) -> bool:
        a_min = obj1.bounding_box[0]
        a_max = obj1.bounding_box[1]
        b_min = obj2.bounding_box[0]
        b_max = obj2.bounding_box[1]
        if a_min[0] >= b_max[0] or b_min[0] >= a_max[0]:
            return False
        if a_min[1] >= b_max[1] or b_min[1] >= a_max[1]:
            return False
        if a_min[2] >= b_max[2] or b_min[2] >= a_max[2]:
            return False
        return True

    def _test(self, obj1: Mesh, obj2: Mesh) -> bool:
        bs_test = self._bounding_sphere_collision(obj1, obj2)
        if not bs_test:
            return False
        if self.level == self.SPHERICAL:
            return True
        if isinstance(obj1, Sphere) and isinstance(obj2, Sphere):
            # If both objects are Spheres, this check is enough
            return True
        if isinstance(obj1, Mesh) and isinstance(obj2, Mesh):
            return self._aabb_collision_test(obj1, obj2)
        return False

    def resolve(self, obj1: Mesh, obj2: Mesh) -> None:
        """Notify the test case for collision resolution.

        Otherwise, tester will ignore the collision in iterations
        to gain more performance

        Order of the objects are not important.

        Keywords:
        obj1 -- Object 1
        obj2 -- Object 2
        """
        pair = set([obj1, obj2])
        if pair in self._pairs:
            self._pairs.remove(pair)

    def check(self):
        """Check the objects for possible collisions"""
        for pair_tuple in combinations(self.objects, 2):
            pair = set(pair_tuple)
            if pair not in self._pairs:
                res = self._test(pair_tuple[0], pair_tuple[1])
                if res:
                    self._pairs.append(pair)

        if len(self._pairs) > 0:
            self.callback(self, [list(pair) for pair in self._pairs])



import threading
import time
from typing import Any, Callable

# SAFE_ASSUMPTION constant is a wait time between each clock cycle.
SAFE_ASSUMPTION = 0.01


class ClockException(Exception):
    pass


class Clock(threading.Thread):
    """Clock thread

    Clock is a time-based thread which triggers the callback function
    in given intervals.

    Some use cases:
    * Get the object states every n microseconds and log them
    * Animate an object (move the object every 1/24 seconds)
    * Create an AI that scans the entire surroumdings regularly

    Best use case is to use it as a global timer in the whole scene
    with a single Clock.

    Due to GIL in Python Interpreter, try to keep the number of clocks as
    low as possible to prevent each clock to get in the way of main
    render cycle. Many clocks would probably create a lag in the scene.

    For example, you can have 10 AI bots in the scene.
    Create a single clock that loops and makes a decision for each bot
    instead of creating separate clocks for each bot.

    """

    def __init__(self, period: float, callback: Callable[[float, float], None]):
        """Initialize the clock

        Keyword arguments:
        period -- Call period of the clock in seconds as FLOAT
        callback -- Callback method to call.
        """
        threading.Thread.__init__(self)
        self.period = period
        self._total_time = 0.0
        self.callback = callback
        self._kill = False
        self._pause = True
        self._hold = False

    def kill(self) -> None:
        """Kill the clock, do not run anymore. (Kill the thread)"""
        self._kill = True

    def pause(self) -> None:
        """Pause the clock temporarily"""
        self._pause = not self._pause

    def run(self) -> Any:
        """Main clock cycle"""
        if self._kill:
            raise ClockException("This clock is already killed")
        while True:
            if self._kill:
                return False

            if self._pause:
                # sleep a safe time to avoid excessive cpu usage
                time.sleep(SAFE_ASSUMPTION)
                continue

            self.callback(self.period, self._total_time)

            time.sleep(self.period)
            self._total_time += self.period
        return True



# pylama:ignore=C901
import ctypes
import logging
from copy import deepcopy
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np  # type: ignore
from OpenGL.GL import (
    GL_ARRAY_BUFFER,
    GL_DEPTH_TEST,
    GL_DYNAMIC_DRAW,
    GL_ELEMENT_ARRAY_BUFFER,
    GL_FILL,
    GL_FLOAT,
    GL_FRONT_AND_BACK,
    GL_LINE,
    GL_LINE_STRIP,
    GL_POINTS,
    GL_TRIANGLES,
    GL_UNSIGNED_INT,
    glBindBuffer,
    glBindVertexArray,
    glBufferData,
    glBufferSubData,
    glDeleteVertexArrays,
    glDisable,
    glDrawElements,
    glEnable,
    glEnableVertexAttribArray,
    glGenBuffers,
    glGenVertexArrays,
    glIsVertexArray,
    glPolygonMode,
    glVertexAttribPointer,
)

from payton.math.functions import (
    add_vectors,
    create_rotation_matrix,
    cross_product,
    normalize_vector,
    scale_matrix,
    scale_vector,
    sub_vector,
    to_4,
    vector_transform,
)
from payton.math.geometry import raycast_sphere_intersect
from payton.math.matrix import Matrix, IDENTITY_MATRIX
from payton.math.vector import Vector2D, Vector3D
from payton.scene.material import DEFAULT, NO_INDICE, NO_VERTEX_ARRAY, POINTS, SOLID, WIREFRAME, Material
from payton.scene.shader import DEFAULT_SHADER, PARTICLE_SHADER, Shader
from payton.scene.types import IList, VList


class Object:
    def __init__(
        self,
        name="",
        visible=True,
        track_motion=False,
        **kwargs: Dict[str, Any],
    ) -> None:
        """Initialize the object

        This is the base Object of Payton and almost all scene objects as well
        as HUD objects.

        Keyword arguments:
        name -- Name of the object
        visible -- Set if the object is visible or not.
        track_motion -- Set only if you really need to track the object's motion path.
                        This comes with an over-head.
        """
        self.children: Dict[str, Object] = {}

        # store diffeerent materials
        self.materials: Dict[str, Material] = {DEFAULT: Material()}

        self._vao: int = NO_VERTEX_ARRAY
        self._vbos: List[int] = []
        self._no_missing_vao = False

        self.name = name
        self._visible = visible
        self.matrix = deepcopy(IDENTITY_MATRIX)
        # Object vertices. Each vertex has 3 decimals (X, Y, Z). Vertices
        # are continuous. [X, Y, Z, X, Y, Z, X, Y, Z, X, ... ]
        #                  -- 1 --  -- 2 --  -- 3 --  -- 4 --

        self._vertices: List[Vector3D] = []  # Object vertex list

        # @NOTE: we have separate indices for materials but this base
        #        index list holds all indice definitions for fast access
        self._indices: IList = []  # Indices
        self._normals: List[Vector3D] = []  # Vertex normals, 1 normal coordinate for 1 Vertex
        self._texcoords: List[Vector2D] = []  # Texture coordinates, 1 coordinate per Vertex
        self._vertex_colors: List[Vector3D] = []  # per-vertex colors, optional.
        self._has_vertex_colors: bool = False  # flag for using vertex colors

        # Vertices do not mean anything unless we define how to use them.
        # For instance, 3 vertices make a triangle or 2 vertices define a line
        # order of vertices are defined in materials indices
        self._vertex_count: int = 0  # Number of vertices to report to OpenGL.

        # This is an optimization technique for dynamic objects where there are
        # increasing number of vertices. We allocate some buffer before-hand
        # and if we fill all of it, we resize it.
        self._buffer_size: float = 500 * 12
        self._t_buffer_size: float = 500 * 8
        self._model_matrix: np.ndarray = []  # Model matrix.
        self._model_matrix_fortran: np.ndarray = []  # Model matrix as fortran
        # Check if buffer size allocated for the object has changed.
        self._buffer_size_changed: bool = True
        self._t_buffer_size_changed: bool = True

        self.shader = DEFAULT_SHADER

        # Track object motion
        self.track_motion = track_motion
        # Motion path, stores every matrix change.
        self._motion_path: List[Matrix] = []

        if not isinstance(self, Line):
            # _motion_path_line is used to display the motion path in scene
            self._motion_path_line = Line()
        self._previous_matrix: Union[np.ndarray, None] = None

        # For raycast tests - bounding radius is the radius of the bounding
        # sphere.
        self._bounding_radius: float = 0
        self._bounding_box: VList = []
        self._selected: bool = False

        # Vertex Array Object pointer
        self._needs_update: bool = False  # Object geometry has changed.
        self._hit: bool = False

        self._absolute_vertices: Optional[List[Vector3D]] = None

    def refresh(self) -> None:
        """Refresh the object

        Object has another memory area in the graphics card. So, for every
        geometrical change in the objects, you need to refresh the object
        so that it gets updated in the graphics card memory.
        """
        self._needs_update = True

    @property
    def material(self) -> Material:
        """Return the DEFAULT material of the object."""
        return self.materials[DEFAULT]

    @material.setter
    def material(self, mat: Material) -> None:
        """Set the DEFAULT material of the object.

        Note: This is a regular python pass-by-reference and doesn't deepcopy
        the actual material properties.

        Keyword arguments:
        mat -- Material to set
        """
        self.materials[DEFAULT] = mat

    def add_material(self, name: str, material: Material) -> None:
        """Add a material to the object.

        An object can have multiple materials for multiple polygons

        Keyword arguments:
        name -- Name of the new material
        material -- Material to set
        """
        if name in self.materials:
            raise Exception(f"Name {name} already exists")
        self.materials[name] = deepcopy(material)

    @property
    def direction(self) -> Vector3D:
        """Get the direction of the Object.

        Ideally, this is the direction that you are facing as a human.
        """
        return self.matrix[1][:3]

    @direction.setter
    def direction(self, v: Vector3D):
        """Set the direction towards a given vector.

        Note: This is a unit vector and be sure to provide a unit vector!
        Otherwise, please use `direct_to` method to point the object to an exact
        position in space.

        Keyword arguments:
        v -- Unit vector to set
        """
        self.matrix = [self.matrix[0], to_4(v, 0), self.matrix[2], self.matrix[3]]

        left = cross_product(self.matrix[1], self.matrix[2])

        self.matrix = [to_4(normalize_vector(left), 0), self.matrix[1], self.matrix[2], self.matrix[3]]

        up = cross_product(self.matrix[0], self.matrix[1])

        self.matrix = [self.matrix[0], self.matrix[1], to_4(up, 0), self.matrix[3]]
        self._to_absolute.cache_clear()
        self._absolute_vertices = None

    def direct_to(self, v: Vector3D):
        """Direct object to given coordinates in space.

        Keyword arguments:
        v -- A point in space. Not a unit vector
        """
        diff = sub_vector(v, self.position)
        diff = normalize_vector(diff)
        self.direction = diff

    def rotate_around_z(self, angle: float) -> None:
        """Rotate the object around Z axis in given angle in Radians

        Keyword arguments:
        angle -- Angle in radians
        """
        rot_matrix = create_rotation_matrix([0, 0, 1], angle)
        local_matrix = np.array(self.matrix, dtype=np.float32)
        local_matrix = rot_matrix.dot(local_matrix)
        self.matrix = local_matrix.tolist()
        self._to_absolute.cache_clear()
        self._absolute_vertices = None

    def rotate_around_x(self, angle: float) -> None:
        """Rotate the object around X axis in given angle in Radians

        Keyword arguments:
        angle -- Angle in radians
        """
        rot_matrix = create_rotation_matrix([1, 0, 0], angle)
        local_matrix = np.array(self.matrix, dtype=np.float32)
        local_matrix = rot_matrix.dot(local_matrix)
        self.matrix = local_matrix.tolist()
        self._to_absolute.cache_clear()
        self._absolute_vertices = None

    def rotate_around_y(self, angle: float) -> None:
        """Rotate the object around Y axis in given angle in Radians

        Keyword arguments:
        angle -- Angle in radians
        """
        rot_matrix = create_rotation_matrix([0, 1, 0], angle)
        local_matrix = np.array(self.matrix, dtype=np.float32)
        local_matrix = rot_matrix.dot(local_matrix)
        self.matrix = local_matrix.tolist()
        self._to_absolute.cache_clear()
        self._absolute_vertices = None

    def scale(self, x: float, y: float, z: float) -> None:
        """Scale the look of the object in the scene.

        This method just applies a scale matrix to the object in space.
        Does not change it's internal coordinates / vertices.

        Keyword arguments:
        x -- Scale in X direction
        y -- Scale in Y direction
        z -- Scale in Z direction
        """
        sm = scale_matrix(x, y, z)
        local_matrix = np.array(self.matrix, dtype=np.float32)
        local_matrix = sm.dot(local_matrix)
        self.matrix = local_matrix.tolist()
        self._to_absolute.cache_clear()
        self._absolute_vertices = None

    def scale_texture(self, x: float, y: float) -> None:
        """Scale texture coordinates of the object.

        Assuming that the object has a stretched texture on it's surface.
        By giving X and Y as > 1, you can repeat the texture on the surface.

        Keyword arguments:
        x -- Scale factor in X
        y -- Scale factor in Y
        """
        self._texcoords = [[coord[0] * x, coord[1] * y] for coord in self._texcoords]
        self.refresh()

    def select(self, start: np.ndarray, vector: np.ndarray) -> bool:
        """Checks if the given ray manages to select the object. Returns boolean.

        This method does a bounding-sphere intersection test as it's the fastest
        intersection test. As a result, this method wouldn't be accurate
        in some cases especially with long and thin objects.

        This is by design and unless an equally fast and more accurate algorithm
        comes to my mind, this will remain the same.

        Keyword arguments:
        start -- Starting position of the ray vector
        vector -- Direction of the ray vector
        """
        self._selected = raycast_sphere_intersect(
            start,
            vector,
            np.array(self.matrix[3], dtype=np.float32),
            self.bounding_radius,
        )

        for obj in self.children:
            x = self.children[obj].select(start, vector)
            if not self._selected and x:
                self._selected = True

        return self._selected

    def destroy(self) -> bool:
        """Destroy the object.

        This method frees the graphics card memory/buffer used for the rendering.
        And sets the virtual pointers to zero to make them re-usable again.
        """
        for material in self.materials.values():
            if material._vao > NO_VERTEX_ARRAY:
                glDeleteVertexArrays(1, [self._vao])
                material._vao = NO_VERTEX_ARRAY
                material._initialized = False

        self._buffer_size_changed = True
        self._t_buffer_size_changed = True
        self._no_missing_vao = False
        if self._vao > NO_VERTEX_ARRAY:
            glDeleteVertexArrays(1, [self._vao])
            self._vao = NO_VERTEX_ARRAY

        for child in self.children.values():
            child.destroy()

        return True

    def step_back(self, steps: int = 1) -> bool:
        """Step back in time if the object is in track_motion mode.

        Note: this is not reversable.

        Keyword arguments:
        steps -- Number of steps to go back in time.
        """
        steps += 1
        if not self.track_motion:
            raise Exception("track_motion should be True")
        if len(self._motion_path) < steps:
            return False

        self.matrix = self._motion_path[-steps]
        self._to_absolute.cache_clear()
        del self._motion_path[-steps + 1 :]  # noqa
        self._absolute_vertices = None
        return True

    def forward(self, distance: float) -> None:
        """Move the object forward by given distance.

        Forward is the object's direction.

        Keyword arguments:
        distance -- Distance to move forward.
        """
        diff = scale_vector(self.matrix[1], distance)
        self.matrix = [self.matrix[0], self.matrix[1], self.matrix[2], to_4(add_vectors(self.matrix[3], diff))]
        self._to_absolute.cache_clear()
        self._absolute_vertices = None

    def update_matrix(self, parent_matrix: Optional[np.ndarray] = None) -> None:
        """Update the objects matrix.

        Object's matrix is stored as Python's native List[List[float]] but
        that type is not suitable for the graphics card.

        Graphics card works with a well-aligned packed byte array whereas Python lists
        have extra bytes for stuff.

        So we turn the python's list into Numpy array, which is compatible to use with
        the graphics card.

        Keyword arguments:
        parent_matrix -- Parent matrix to multiply our initial matrix. (Optional)
        """
        # Turn matrix into numpy array. Numpy arrays are C Type arrays
        # suitable for OpenGL Pipeline
        self._model_matrix = np.array(self.matrix, dtype=np.float32)

        # When there is a parent object, child object follows parents matrix
        if parent_matrix is not None and len(parent_matrix) > 0:
            self._model_matrix = self._model_matrix.dot(parent_matrix)

        # Fortran array is used to push matrix to opengl context
        self._model_matrix_fortran = np.asfortranarray(self._model_matrix, dtype=np.float32)

    def track(self) -> bool:
        """Track the object movement if it has changed from the previous frame"""
        if not self.track_motion:
            return False

        if self._previous_matrix == self.matrix[3]:
            return True

        # Add the new matrix to motion path records
        self._motion_path.append(deepcopy(self.matrix))
        # Add the matrix position to motion math line for visualisation
        if self._motion_path_line is not None:
            self._motion_path_line.append([[self.matrix[3][0], self.matrix[3][1], self.matrix[3][2], 1.0]])

        # Python trick here! need to .copy or it will pass reference.
        self._previous_matrix = self.matrix[3]
        return True

    @property
    def visible(self) -> bool:
        """Return if the object is visible"""
        return self._visible

    def show(self) -> None:
        """Show the object (set visible = True)"""
        self._visible = True

    def hide(self) -> None:
        """Hide the object (set visible = False)"""
        self._visible = False

    def clone(self) -> "Object":
        """Create the clone of an object - deepcopy

        Caution! This will copy literally everything, that includes the OpenGL buffer pointers
        so changing one of the objects will result with changing the same memory area
        in graphics card!
        """
        return deepcopy(self)

    @property
    def has_missing_vao(self) -> bool:
        """Check if the object is missing any Virtual Array Objects. Return boolean"""
        if self._no_missing_vao:
            return False
        result = any(
            [material._vao == NO_VERTEX_ARRAY and len(material._indices) > 1 for material in self.materials.values()]
        )
        if not result:
            self._no_missing_vao = True
        return result

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: int = None,
    ) -> None:
        """Main render cycle of the object.

        Keyword arguments:
        lit -- Is the object illuminated?
        shader -- Shader to use while rendering the object
        parent_matrix -- Parent object's matrix if this is a child object
        _primitive -- override the scene wide primitive (rendering) mode. (Point, Wire, Solid)
        """
        if not self._visible:
            return

        if self.has_missing_vao or self._needs_update:
            self.build()

        self.update_matrix(parent_matrix=parent_matrix)
        self.track()

        if self._vertex_count == 0:
            # dummy object, render children and leave
            # render children
            for child in self.children:
                self.children[child].render(
                    lit,
                    shader,
                    self._model_matrix,
                )

            return

        # Material shading mode.
        mode = None
        if self._has_vertex_colors:
            mode = Shader.PER_VERTEX_COLOR

        shader.set_matrix4x4_np("model", self._model_matrix_fortran)
        indice_0 = ctypes.c_void_p(0)

        if self.shader == PARTICLE_SHADER:
            glDisable(GL_DEPTH_TEST)

        for material in self.materials.values():
            if not material.display == SOLID and shader._depth_shader:
                continue

            material.render(
                lit,
                shader,
                mode,
            )

            # Actual rendering
            if material._vao > NO_VERTEX_ARRAY and glIsVertexArray(material._vao):
                glBindVertexArray(material._vao)
                pmode = GL_LINE
                primitive = GL_TRIANGLES
                if material.display == SOLID:
                    pmode = GL_FILL
                    primitive = GL_TRIANGLES
                if material.display == POINTS:
                    pmode = GL_FILL
                    primitive = GL_POINTS
                if _primitive is not None:
                    primitive = _primitive
                glPolygonMode(GL_FRONT_AND_BACK, pmode)

                glDrawElements(
                    primitive,
                    material._vertex_count,
                    GL_UNSIGNED_INT,
                    indice_0,
                )

                if pmode != GL_FILL:
                    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
                glBindVertexArray(0)

        if self.shader == PARTICLE_SHADER:
            glEnable(GL_DEPTH_TEST)

        # Render motion path
        if self.track_motion:
            self._motion_path_line.render(lit, shader, parent_matrix)

        # render children
        for child in self.children:
            self.children[child].render(lit, shader, self._model_matrix)

    def set_position(self, x: float, y: float, z: float):
        """Set the position of the object

        Keyword arguments:
        x -- X Position of the object
        y -- Y Position of the object
        z -- Z Position of the object
        """
        self.matrix[3] = [x, y, z, 1.0]
        self._to_absolute.cache_clear()

    @property
    def position(self) -> Vector3D:
        """Return the object's position"""
        return self.matrix[3][:3]

    @position.setter
    def position(self, pos: Union[Vector3D, list, tuple]) -> None:
        """Set the position of the object by a given Vector

        In this context, the definition of the position is extended to
        support List[float] and Tuple as well to give some space to developers

        Note: This position is relative to it's parent.

        Keyword arguments:
        pos -- Position to set
        """
        fpos = to_4(pos, 1.0)
        self.matrix[3] = fpos
        self._absolute_vertices = None
        self._to_absolute.cache_clear()

    def add_child(self, name: str, obj: "Object") -> bool:
        """Add a child object to the object.

        Child objects follow their parent's movements, visibility and etc.
        Think of them as your arms and legs.

        Keyword arguments:
        name -- Name of the child object (must be unique within the same object)
        obj -- Object to add
        """
        if name in self.children:
            logging.error(f"Name {name} exists in object children")
            return False
        if not isinstance(obj, Object):
            logging.error("Object type is not valid")
            return False
        self.children[name] = obj
        return True

    @lru_cache(maxsize=10240)
    def _to_absolute(self, coordinate: Tuple) -> Vector3D:
        return vector_transform(list(coordinate), self.matrix)

    def to_absolute(self, coordinate: Vector3D) -> Vector3D:
        """Convert the given coordinates to absolute space coordinates. Returns the converted vector.

        Child object's coordinates are relative to it's parent.
        You can use this method if you want to know the absoltue (root) coordinates of a local coordinate.

        Keyword arguments:
        coordinate -- Coordinate to convert"""

        return self._to_absolute(tuple(coordinate))

    def absolute_vertices(self) -> List[Vector3D]:
        """Convert all object vertices into absolute vertices.

        This is a cpu intesive operation and should be done with caution.
        Detailed collision detection calls this method.

        Sets `self._absolute_vertices` if None, else returns it as is.
        """
        if self._absolute_vertices is None:
            self._absolute_vertices = [self._to_absolute(tuple(v)) for v in self._vertices]

        return self._absolute_vertices

    def toggle_wireframe(self) -> None:
        """Toggle the material display type from Wireframe / Particle / Solid"""
        d = (self.material.display + 1) % 3

        for mat in self.materials.values():
            mat.display = d

        if d == POINTS:
            self.shader = PARTICLE_SHADER
            self.refresh()
        else:
            self.shader = DEFAULT_SHADER
            self.refresh()

        for child in self.children.values():
            child.toggle_wireframe()

    def _calc_bounds(self) -> float:
        bmin: Optional[Vector3D] = None
        bmax: Optional[Vector3D] = None
        absolute_vertices = self.absolute_vertices()
        x = [v[0] for v in absolute_vertices]
        y = [v[1] for v in absolute_vertices]
        z = [v[2] for v in absolute_vertices]

        bmin = [0.0, 0.0, 0.0]
        bmax = [0.0, 0.0, 0.0]
        if len(x) != 0 and len(y) != 0 and len(z) != 0:
            bmin = [min(x), min(y), min(z)]
            bmax = [max(x), max(y), max(z)]

        a = (bmax[0] - bmin[0]) / 2.0
        b = (bmax[1] - bmin[1]) / 2.0
        c = (bmax[2] - bmin[2]) / 2.0
        self._bounding_radius = max([a, b, c])
        self._bounding_box = [bmin, bmax]

        return self._bounding_radius

    @property
    def bounding_radius(self) -> float:
        """Get the bounding sphere radius"""
        if self._bounding_radius > 0:
            return self._bounding_radius
        return self._calc_bounds()

    @property
    def bounding_box(self) -> VList:
        """Get the bounding box (AABB) coordinates"""
        if self._absolute_vertices is None:
            self._calc_bounds()
        if len(self._bounding_box) > 0:
            return self._bounding_box
        self._calc_bounds()
        return self._bounding_box

    def build(self) -> bool:
        """Build the object so that it can be rendered in pipeline.

        Conver object vertices / normals / texture coordinates / vertex colors
        into Numpy arrays. Generate the Vertex Arrays and Buffer for the object
        in Graphics Pipeline. Load the data into graphics memory.
        Set the relevant pointer informations.
        """
        self._vertex_count = 0

        # Turn python arrays into C type arrays using Numpy.
        # This is required for OpenGL. Python memory model is a bit
        # different than raw memory model of C (OpenGL)
        vertices = np.array(self._vertices, dtype=np.float32).flatten()
        normals = np.array(self._normals, dtype=np.float32).flatten()
        texcoords = np.array(self._texcoords, dtype=np.float32).flatten()
        colors = np.array(self._vertex_colors, dtype=np.float32).flatten()

        if self._vao == NO_INDICE:
            return False

        if self._vao == NO_VERTEX_ARRAY:
            self._vao = glGenVertexArrays(1)
            self._vbos = glGenBuffers(5)

        # We will let all materials to share the same buffer objects
        # We need 4 buffers (vertex, normal, texcoord, color)

        glBindVertexArray(self._vao)

        draw = GL_DYNAMIC_DRAW

        # Buffer overflow, we need more space.
        if self._buffer_size < vertices.nbytes:
            self._buffer_size = vertices.nbytes
            self._buffer_size_changed = True
        if self._t_buffer_size < texcoords.nbytes:
            self._t_buffer_size = texcoords.nbytes
            self._t_buffer_size_changed = True

        # Bind Vertices
        glBindBuffer(GL_ARRAY_BUFFER, self._vbos[0])
        if self._buffer_size_changed:
            # glBufferData creates a new data area
            glBufferData(GL_ARRAY_BUFFER, self._buffer_size, vertices, draw)
        else:
            # glBufferSubData just replaces memory area in buffer so it is
            # much more efficient way to handle things.
            glBufferSubData(GL_ARRAY_BUFFER, 0, vertices.nbytes, vertices)

        # Bind Normals
        glBindBuffer(GL_ARRAY_BUFFER, self._vbos[1])
        if self._buffer_size_changed:
            glBufferData(GL_ARRAY_BUFFER, self._buffer_size, normals, draw)
        else:
            glBufferSubData(GL_ARRAY_BUFFER, 0, normals.nbytes, normals)

        # Bind TexCoords
        if len(self._texcoords) == len(self._vertices):
            glBindBuffer(GL_ARRAY_BUFFER, self._vbos[2])
            if self._t_buffer_size_changed:
                glBufferData(GL_ARRAY_BUFFER, self._t_buffer_size, texcoords, draw)
            else:
                glBufferSubData(GL_ARRAY_BUFFER, 0, texcoords.nbytes, texcoords)

        # Bind Vertex Colors
        if len(self._vertex_colors) == len(self._vertices):
            glBindBuffer(GL_ARRAY_BUFFER, self._vbos[4])
            self._has_vertex_colors = True
            if self._buffer_size_changed:
                glBufferData(GL_ARRAY_BUFFER, self._buffer_size, colors, draw)
            else:
                glBufferSubData(GL_ARRAY_BUFFER, 0, colors.nbytes, colors)

        glBindVertexArray(0)
        self._buffer_size_changed = False
        self._t_buffer_size_changed = False

        for material in self.materials.values():
            if len(material._indices) == 0:
                material._vao == NO_INDICE
                continue

            if material._vao == NO_VERTEX_ARRAY:
                # Generate Vertex Array
                material._vao = glGenVertexArrays(1)
                # We need 1 buffer for material as indices
                material._vbos = [glGenBuffers(1)]

            glBindVertexArray(material._vao)

            indices = np.array(material._indices, dtype=np.int32).flatten()

            # Bind Vertices
            glBindBuffer(GL_ARRAY_BUFFER, self._vbos[0])
            glEnableVertexAttribArray(0)  # shader layout location
            glVertexAttribPointer(0, 3, GL_FLOAT, False, 0, ctypes.c_void_p(0))

            # Bind Normals
            glBindBuffer(GL_ARRAY_BUFFER, self._vbos[1])
            glEnableVertexAttribArray(1)  # shader layout location
            glVertexAttribPointer(1, 3, GL_FLOAT, False, 0, ctypes.c_void_p(0))

            # Bind TexCoords
            if len(self._texcoords) == len(self._vertices):
                glBindBuffer(GL_ARRAY_BUFFER, self._vbos[2])
                glEnableVertexAttribArray(2)  # shader layout location
                glVertexAttribPointer(2, 2, GL_FLOAT, False, 0, ctypes.c_void_p(0))

            # Bind Vertex Colors
            if len(self._vertex_colors) == len(self._vertices):
                glBindBuffer(GL_ARRAY_BUFFER, self._vbos[4])
                glEnableVertexAttribArray(3)  # shader layout location
                glVertexAttribPointer(3, 3, GL_FLOAT, False, 0, ctypes.c_void_p(0))

            # Bind Indices
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, material._vbos[0])
            glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.nbytes, indices, draw)
            i_len = len(indices)
            material._vertex_count = i_len
            self._vertex_count += i_len

            glBindVertexArray(0)
            glBindBuffer(GL_ARRAY_BUFFER, 0)

        self._needs_update = False
        return True

    def to_dict(self) -> Dict[str, Any]:
        """Convert the object into a dictionary for export / debug"""
        return {
            "vertices": self._vertices,
            "normals": self._normals,
            "texcoords": self._texcoords,
            "matrix": self.matrix,
            "materials": {mat: self.materials[mat].to_dict() for mat in self.materials},
            "children": {n: self.children[n].to_dict() for n in self.children},
        }


class Line(Object):
    def __init__(
        self,
        vertices: Optional[VList] = None,
        color: Optional[Vector3D] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize a Line object

        Keyword arguments:
        vertices -- Optional Vertices information for the line
        color -- Optional color of the line material.
        """
        super().__init__(**kwargs)
        self._vertices: VList = [] if vertices is None else vertices
        self.material.color = [1.0, 1.0, 1.0] if color is None else color

        self.material.display = WIREFRAME
        self.material.lights = False
        self.build_lines()

    def toggle_wireframe(self) -> None:
        """Toggle Wireframe overwrite to disable mode change"""

    def add_material(self, name: str, material: Material) -> None:
        """@TODO Implement this later! Not urgent"""
        logging.error("Can't add materials to Line object")

    def render(
        self,
        lit: bool,
        shader: Shader,
        parent_matrix: Optional[np.ndarray] = None,
        _primitive: int = None,
    ) -> None:
        """Same as the Object Render but with explicitly defined OpenGL primitive type"""
        super().render(lit, shader, parent_matrix, GL_LINE_STRIP)

    def append(self, vertices: VList, material: str = DEFAULT) -> None:
        """Append vertices to the line

        Keyword arguments:
        vertices -- List of vertices to append
        material -- Material name to append lines into"""
        diff = len(vertices)  # Number of vertices added
        last_index = len(self._vertices)
        self._vertices += vertices

        self._texcoords += [[0, 0]] * diff
        self._normals += [[0, 0, 1]] * diff
        self._vertex_count = len(self._vertices)
        self.material._vertex_count = self._vertex_count

        indices = list(map(lambda x: x + last_index, range(diff)))
        self.material._indices.extend([indices])
        self._indices = self.material._indices

        self._needs_update = True

    def build_lines(self, vertices: Optional[VList] = None, color: Optional[Vector3D] = None) -> None:
        """Build lines information

        Keyword arguments:
        vertices -- (Optional) if you want to reset the whole line vertices in one go.
        color -- Color of the DEFAULT material
        """
        if vertices is not None:
            self._vertices = vertices
        if color is not None:
            self.material.color = color
        self._vertex_count = len(self._vertices)
        self.material._vertex_count = self._vertex_count

        for i in range(self._vertex_count - 1):
            self.material._indices.append([i, i + 1])
            self._indices = self.material._indices

        for i in range(self._vertex_count):
            self._normals.append([0, 0, 0])
            self._texcoords.append([0, 0])
        self._needs_update = True



from typing import Union

import numpy as np  # type: ignore
from payton.math.vector import Vector3D
from payton.math.matrix import Matrix

GArray = Union[Vector3D, np.ndarray]
GMatrix = Union[Matrix, np.ndarray]



from typing import Any, Callable, List, Tuple
from payton.math.vector import Vector3D

VList = List[Vector3D]
VListW = List[Vector3D]
IList = List[List[int]]
CPlane = Tuple[Vector3D, Vector3D, Callable[[Vector3D], Any]]



# pylama:ignore=C901
import logging
from typing import Any, List, Optional

import sdl2

from payton.scene.gui import EditBox
from payton.scene.camera import BUTTON_LEFT, BUTTON_MIDDLE, BUTTON_RIGHT


class BaseController:
    def __init__(self) -> None:
        super().__init__()

    def keyboard(self, event: sdl2.SDL_Event, scene: Any) -> bool:
        """
        Returns boolean to indicate if this controller interrupts the chain
        """
        raise NotImplementedError()

    def mouse(self, event: sdl2.SDL_Event, scene: Any) -> bool:
        """
        Returns boolean to indicate if this controller interrupts the chain
        """
        raise NotImplementedError()


class GUIController(BaseController):
    """
    Define GUI related controller actions in this class
    to reduce complexity inside Controller.
    """

    def __init__(self) -> None:
        super().__init__()
        self._active_object: Optional[EditBox] = None

    def keyboard(self, event: sdl2.SDL_Event, scene: Any) -> bool:
        """Keyboard handler for the base GUI controls. Returns True if keyboard action
        stops the event check cycle.

        Keyword arguments:
        event -- SDL2 Event Triggered
        scene -- Active scene
        """
        if event.type == sdl2.SDL_TEXTINPUT:
            if self._active_object is not None:
                self._active_object._on_keypress(event.text.text.decode('utf-8'))
                return True

        if event.type == sdl2.SDL_KEYUP:
            key = event.key.keysym.sym
            if self._active_object is not None:
                if key == sdl2.SDLK_ESCAPE or (key == sdl2.SDLK_RETURN and not self._active_object.multiline):
                    self._active_object._exit()
                    self._active_object = None
                    sdl2.SDL_ShowCursor(True)
                    sdl2.SDL_StopTextInput()
                if self._active_object and key == sdl2.SDLK_BACKSPACE:
                    self._active_object.backspace()
                if key == sdl2.SDLK_RETURN and self._active_object is not None and self._active_object.multiline:
                    self._active_object._on_keypress("\n")
                if self._active_object and key == sdl2.SDLK_LEFT:
                    self._active_object.cursor_left()
                if self._active_object and key == sdl2.SDLK_RIGHT:
                    self._active_object.cursor_right()
                return True
        return False

    def mouse(self, event: sdl2.SDL_Event, scene: Any) -> bool:
        """Mouse handler for the base GUI controls. Returns True if the mouse action stops the
        event check cycle.

        Keyword arguments:
        event -- SDL2 Event Triggered
        scene -- Active scene
        """
        if event.type == sdl2.SDL_MOUSEBUTTONDOWN:
            mx, my = event.button.x, event.button.y
            for hud in scene.huds:
                h = scene.huds[hud]
                for shape in h.children:
                    focus_element = h.children[shape].click(mx, my)
                    if focus_element:
                        if self._active_object is not None and self._active_object != focus_element:
                            self._active_object._exit()
                            self._active_object = None
                            sdl2.SDL_ShowCursor(True)
                            sdl2.SDL_StopTextInput()
                        if hasattr(focus_element, "_on_keypress"):
                            self._active_object = focus_element
                            sdl2.SDL_ShowCursor(False)
                            sdl2.SDL_StartTextInput()
                        return True
        return False


class SceneController(BaseController):
    """SceneController controls the camera actions along with clocks and other
    in-scene actions
    """

    def keyboard(self, event: sdl2.SDL_Event, scene: Any) -> bool:
        """Keyboard handler for the whole scene

        Keyword arguments:
        event -- Triggered SDL2 Event
        scene -- Active scene
        """
        if event.type == sdl2.SDL_QUIT:
            logging.debug("Quit SDL Scene")
            scene.terminate()
            return True

        if event.type == sdl2.SDL_KEYDOWN:
            key = event.key.keysym.sym
            if key == sdl2.SDLK_LSHIFT:
                scene._shift_down = True
            if key == sdl2.SDLK_LCTRL:
                scene._ctrl_down = True

        if event.type == sdl2.SDL_KEYUP:
            key = event.key.keysym.sym

            if key == sdl2.SDLK_LSHIFT:
                scene._shift_down = False
            if key == sdl2.SDLK_LCTRL:
                scene._ctrl_down = False

            if key == sdl2.SDLK_c:
                # below variable assignment is only for code style
                p = scene.active_camera.perspective
                scene.active_camera.perspective = not p
                logging.debug(f"Camera Perspective={scene.active_camera.perspective}")

            if key == sdl2.SDLK_g:
                scene.grid.visible = not scene.grid.visible

            if key == sdl2.SDLK_SPACE:
                for clock in scene.clocks:
                    c = scene.clocks[clock]
                    logging.debug(f"Pause clock [{clock}]")
                    c.pause()

            if key == sdl2.SDLK_h:
                scene.huds["_help"]._visible = not scene.huds["_help"]._visible

            if key == sdl2.SDLK_w:
                for obj in scene.objects:
                    scene.objects[obj].toggle_wireframe()

            if key == sdl2.SDLK_ESCAPE:
                scene.running = False

            if key in [sdl2.SDLK_F2, sdl2.SDLK_F3]:
                active = 0
                for i in range(len(scene.cameras)):
                    if scene.cameras[i].active:
                        active = i
                if key == sdl2.SDLK_F2:
                    active -= 1
                if key == sdl2.SDLK_F3:
                    active += 1
                if active < 0:
                    active = len(scene.cameras) - 1
                if active > len(scene.cameras) - 1:
                    active = 0
                scene.active_camera = scene.cameras[active]
                for i in range(len(scene.cameras)):
                    scene.cameras[i].active = i == active
        return False

    def mouse(self, event: sdl2.SDL_Event, scene: Any) -> bool:
        """Mouse handler for the scene

        Keyword arguments:
        event -- Triggered SDL2 Event
        scene -- Active scene
        """
        camera = scene.active_camera
        if event.type == sdl2.SDL_MOUSEBUTTONUP:
            camera._prev_intersection = None

        if event.type == sdl2.SDL_MOUSEBUTTONDOWN:
            mx, my = event.button.x, event.button.y
            eye, ray_dir = camera.screen_to_world(mx, my, scene.window_width, scene.window_height)

            list = []
            if callable(scene.on_select):
                for obj in scene.objects:
                    hit = scene.objects[obj].select(eye, ray_dir)
                    if hit:
                        list.append(scene.objects[obj])
                if len(list) > 0:
                    scene.on_select(list)

            if not (scene._shift_down or scene._ctrl_down):
                scene._check_click_plane(eye, ray_dir)

        if event.type == sdl2.SDL_MOUSEWHEEL:
            yrel = event.wheel.y
            camera.mouse_wheel(yrel)

        if event.type == sdl2.SDL_MOUSEMOTION:
            button = -1
            if event.motion.state == sdl2.SDL_BUTTON_LMASK:
                button = BUTTON_LEFT
            if event.motion.state == sdl2.SDL_BUTTON_RMASK:
                button = BUTTON_RIGHT
            if event.motion.state == sdl2.SDL_BUTTON_MIDDLE:
                button = BUTTON_MIDDLE
            camera.mouse_move(
                button,
                scene._shift_down,
                scene._ctrl_down,
                event.motion.x,
                event.motion.y,
                event.motion.xrel,
                event.motion.yrel,
                scene.window_width,
                scene.window_height,
            )
        return False


class Controller:
    """Base controller class"""

    def __init__(self):
        super().__init__()
        self._controllers: List[BaseController] = []

    def add_controller(self, controller: BaseController) -> None:
        self._controllers.append(controller)

    def keyboard(self, event: sdl2.SDL_Event, scene: Any) -> None:
        for controller in self._controllers:
            stop_action = controller.keyboard(event, scene)
            if stop_action:
                break

    def mouse(self, event: sdl2.SDL_Event, scene: Any) -> None:
        for controller in self._controllers:
            stop_action = controller.mouse(event, scene)
            if stop_action:
                break
